#ifndef MSP_CORE_REFPTR_H_
#define MSP_CORE_REFPTR_H_

namespace Msp {

struct RefCounts
{
	enum
	{
		KEEP = 1U<<(sizeof(unsigned)*8-1)
	};

	unsigned count;

	RefCounts(): count(0) { }
};


/**
A reference counting smart pointer.  When the last RefPtr for the data gets
destroyed, the data is deleted as well.
*/
template<typename T>
class RefPtr
{
	template<typename U> friend class RefPtr;

private:
	T *data;
	RefCounts *counts;

public:
	RefPtr(): data(0), counts(0) { }
	RefPtr(T *d): data(d), counts(data ? new RefCounts : 0) { incref(); }
private:
	RefPtr(T *d, RefCounts *c): data(d), counts(d ? c : 0) { incref(); }

public:
	/* Must have this or the compiler will generate a default copy-c'tor despite
	the template version */
	RefPtr(const RefPtr &p): data(p.data), counts(p.counts) { incref(); }

	template<typename U>
	RefPtr(const RefPtr<U> &p): data(p.data), counts(p.counts) { incref(); }

	~RefPtr() { decref(); }

	RefPtr &operator=(T *d)
	{
		decref();
		data = d;
		counts = (d ? new RefCounts : 0);
		incref();
		return *this;
	}

	// Likewise for the assignment operator
	RefPtr &operator=(const RefPtr &p) { return assign(p); }

	template<typename U>
	RefPtr &operator=(const RefPtr<U> &p) { return assign(p); }

private:
	template<typename U>
	RefPtr &assign(const RefPtr<U> &p)
	{
		decref();
		data = p.data;
		counts = p.counts;
		incref();
		return *this;
	}

public:
	/** Makes the RefPtr release its reference of the data without deleting it.
	Note that if there are other RefPtrs left with the same data, it might
	still get deleted automatically. */
	T *release()
	{
		T *d = data;
		data = 0;
		decref();
		counts = 0;
		return d;
	}

	/** Marks the data to not be deleted.  This affects all RefPtrs with the
	same data. */
	void keep()
	{
		if(counts)
			counts->count |= RefCounts::KEEP;
	}

	T *get() const { return data; }
	T &operator*() const { return *data; }
	T *operator->() const { return data; }
	operator bool() const { return data!=0; }

	unsigned refcount() const { return (data ? counts->count : 0); }

	template<typename U>
	static RefPtr<T> cast_dynamic(const RefPtr<U> &p)
	{ return RefPtr<T>(dynamic_cast<T *>(p.data), p.counts); }

private:
	void incref()
	{
		if(!counts) return;
		++counts->count;
	}

	void decref()
	{
		if(!counts) return;
		--counts->count;
		if(!counts->count)
		{
			delete data;
			delete counts;
			data = 0;
			counts = 0;
		}
		else if(counts->count==RefCounts::KEEP)
		{
			delete counts;
			data = 0;
			counts = 0;
		}
	}
};

} // namespace Msp

#endif
